# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from typing import List

sys.path.insert(0, os.path.abspath("../"))

# -- Project information -----------------------------------------------------

project = "fedcloudclient"
copyright = "2022, Viet Tran"
author = "Viet Tran"

# The full version, including alpha/beta/rc tags
# release = '0.0.2-dev15'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
]

smartquotes = False

source_suffix = [".rst", ".md"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns: List[str] = []

# -- Options for HTML output -------------------------------------------------

html_context = {
    "display_github": True,
    "github_user": "tdviet",
    "github_repo": "fedcloudclient",
    "github_version": "master/docs/",
}

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_logo = "fedcloudclient-logo-non-transparent-small.png"
html_theme_options = {
    "logo_only": True,
    "display_version": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



"""
"fedcloud site" commands will read site configurations and manipulate with them. If
the local site configurations exist at ~/.config/fedcloud/site-config/, fedcloud will
read them from there, otherwise the commands will read from GitHub repository.

By default, fedcloud does not save anything on local disk, users have to save the
site configuration to local disk explicitly via "fedcloud site save-config" command.
The advantage of having local site configurations, beside faster loading, is to give
users ability to make customizations, e.g. add additional VOs, remove sites they
do not have access, and so on.
"""

import builtins
import json
import shutil
from pathlib import Path
from typing import List
from urllib.request import Request, urlopen

import click
import pkg_resources
import yaml
from jsonschema import validate

from fedcloudclient.decorators import (
    ALL_SITES_KEYWORDS,
    DEFAULT_PROTOCOL,
    all_site_params,
    oidc_params,
    site_vo_params,
)
from fedcloudclient.shell import print_set_env_command

__REMOTE_CONFIG_FILE = (
    "https://raw.githubusercontent.com/tdviet/fedcloudclient/master/config/sites.yaml"
)

# WARNING:
# ALL FILES in this site-config directory WILL BE DELETED  at every execution
# of "fedcloud site save-config" command. Do not change it to a common directory
__LOCAL_CONFIG_DIR = ".config/fedcloud/site-config/"

__site_config_data: List[dict] = []

__FILE_SIZE_LIMIT = 1024 * 1024  # Max size for config files


def read_site_schema():
    """
    Read schema.json for validating site configuration

    :return: JSON object from schema.json
    """
    file = pkg_resources.resource_stream("fedcloudclient", "schema.json")
    schema = json.load(file)
    return schema


def read_site_config():
    """
    Read site configurations from local config dir if exist, otherwise from default
    GitHub location. Storing site configurations in global variable, that will be
    used by other functions. Call read_local_site_config()
    or read_default_site_config()

    :return: None
    """
    if len(__site_config_data) > 0:
        return
    config_dir = Path.home() / __LOCAL_CONFIG_DIR
    if config_dir.exists():
        read_local_site_config(config_dir)
    else:
        read_default_site_config()


def safe_read_yaml_from_url(url, max_length):
    """
    Safe reading from URL
    Check URL and size before reading

    :param url:
    :param max_length:
    :return: data from URL
    """
    if isinstance(url, str) and url.lower().startswith(
        "https://"
    ):  # Only read from HTTPS location
        req = Request(url)
    else:
        raise SystemExit(f"Error: remote filename not starting with https:// : {url}")

    # URLs already checked, so ignore bandit test
    data = None
    try:
        with urlopen(req) as yaml_file:  # nosec
            if int(yaml_file.headers["Content-Length"]) > max_length:
                raise SystemExit(
                    f"Error: remote file {url} is larger than limit {max_length} "
                )
            data = yaml.safe_load(yaml_file)
    except Exception as exception:
        print(f"Error during reading data from {url}")
        raise SystemExit(f"Exception: {exception}")
    return data


def read_default_site_config():
    """
    Read default site configurations from GitHub
    Storing site configurations in a global variable that will be used by other functions

    :return: None
    """
    __site_config_data.clear()
    schema = read_site_schema()

    site_list = safe_read_yaml_from_url(__REMOTE_CONFIG_FILE, __FILE_SIZE_LIMIT)

    # Check data format of site list
    if not isinstance(site_list, builtins.list):
        raise SystemExit("Error: site list is in a wrong format")

    for filename in site_list:
        site_info = safe_read_yaml_from_url(filename, __FILE_SIZE_LIMIT)

        # Validating site config after reading
        try:
            validate(instance=site_info, schema=schema)
        except Exception as exception:
            print(f"Site config in file {filename} is in wrong format")
            raise SystemExit(f"Exception: {exception}")

        __site_config_data.append(site_info)


def read_local_site_config(config_dir):
    """
    Read site configurations from local directory specified in config_dir
    Storing site configurations in global variable, that will be used by other functions

    :param config_dir: path to directory containing site configuration
    :return: None
    """
    __site_config_data.clear()
    schema = read_site_schema()
    config_dir = Path(config_dir)
    for file in sorted(config_dir.glob("*.yaml")):
        try:
            yaml_file = file.open()
            site_info = yaml.safe_load(yaml_file)
            validate(instance=site_info, schema=schema)
            __site_config_data.append(site_info)
        except Exception as exception:
            print(f"Error during reading site config from {file}")
            raise SystemExit(f"Exception: {exception}")


def save_site_config(config_dir):
    """
    Save site configs to local directory specified in config_dir
    Overwrite local configs if exist

    :param config_dir: path to directory containing site configuration
    :return: None
    """
    config_dir = Path(config_dir)
    config_dir.mkdir(parents=True, exist_ok=True)
    for site_info in __site_config_data:
        config_file = config_dir / (site_info["gocdb"] + ".yaml")
        with config_file.open("w", encoding="utf-8") as file:
            yaml.dump(site_info, file)


def delete_site_config(config_dir):
    """
    Delete site configs to local directory specified in config_dir

    :param config_dir: path to directory containing site configuration
    :return: None
    """
    shutil.rmtree(config_dir, ignore_errors=True)


def list_sites():
    """
    List of all sites IDs in site configurations

    :return: list of site IDs
    """
    read_site_config()
    result = []
    for site_info in __site_config_data:
        result.append(site_info["gocdb"])
    return result


def find_site_data(site_name):
    """
    Return configuration of the correspondent site with site_name

    :param site_name: site ID in GOCDB
    :return: configuration of site if found, otherwise None
    """
    read_site_config()

    for site_info in __site_config_data:
        if site_info["gocdb"] == site_name:
            return site_info
    return None


def find_endpoint_and_project_id(site_name, vo):
    """
    Return Keystone endpoint and project ID from site name and VO according
    to site configuration

    :param site_name: site ID in GOCDB
    :param vo: VO name or None to find site endpoint only
    :return: endpoint, project_id, protocol if the VO has access to the site,
             otherwise None, None, None
    """
    site_info = find_site_data(site_name)
    if site_info is None:
        return None, None, None

    protocol = site_info.get("protocol")
    # If only site name is given, return endpoint without project ID
    if vo is None:
        return site_info["endpoint"], None, protocol

    for vo_info in site_info["vos"]:
        if vo_info["name"] == vo:
            return site_info["endpoint"], vo_info["auth"]["project_id"], protocol

    # Return None if VO not found among those that have access to site
    return None, None, None


def find_vo_from_project_id(site_name, project_id):
    """
    Return the VO name form the project ID and site_name according
    to site configuration

    :param site_name: site ID in GOCDB
    :param project_id: project_id configured to support the VO
    :return: vo if the VO is configured, otherwise None
    """
    site_info = find_site_data(site_name)
    if site_info is None:
        return None

    for vo_info in site_info["vos"]:
        if vo_info.get("auth", {}).get("project_id", None) == project_id:
            return vo_info["name"]

    # Return None if the project is not found
    return None


@click.group()
def site():
    """
    Obtain site configurations
    """


@site.command()
@all_site_params
def show(site, all_sites):
    """
    Print configuration of specified site(s)
    """
    if site in ALL_SITES_KEYWORDS or all_sites:
        read_site_config()
        for site_info in __site_config_data:
            site_info_str = yaml.dump(site_info, sort_keys=True)
            print(site_info_str)

    else:
        site_info = find_site_data(site)
        if site_info:
            print(yaml.dump(site_info, sort_keys=True))
        else:
            raise SystemExit(f"Site {site} not found")


@site.command()
@site_vo_params
def show_project_id(site, vo):
    """
    Print Keystone endpoint and project ID
    """
    if site in ALL_SITES_KEYWORDS:
        print("Cannot get project ID for ALL_SITES")
        raise click.Abort()

    endpoint, project_id, _ = find_endpoint_and_project_id(site, vo)
    if endpoint:
        print_set_env_command("OS_AUTH_URL", endpoint)
        print_set_env_command("OS_PROJECT_ID", project_id)
    else:
        raise SystemExit(f"VO {vo} not found on site {site}")


@site.command()
def save_config():
    """
    Load site configs from GitHub, save them to local folder in $HOME
    Overwrite local configs if exist.
    """
    read_default_site_config()
    config_dir = Path.home() / __LOCAL_CONFIG_DIR
    print(f"Saving site configs to directory {config_dir}")
    delete_site_config(config_dir)
    save_site_config(config_dir)


@site.command("list")
def list_():
    """
    List all sites
    """
    read_site_config()
    for site_info in __site_config_data:
        print(site_info["gocdb"])


@site.command()
@site_vo_params
@oidc_params
def env(
    site,
    vo,
    access_token,
):
    """
    Generate OS environment variables for site.
    May set also environment variable OS_ACCESS_TOKEN,
    if access token is provided, otherwise print notification
    """
    if site in ALL_SITES_KEYWORDS:
        print("Cannot generate environment variables for ALL_SITES")
        raise click.Abort()

    endpoint, project_id, protocol = find_endpoint_and_project_id(site, vo)
    if endpoint:
        if protocol is None:
            protocol = DEFAULT_PROTOCOL
        print_set_env_command("OS_AUTH_URL", endpoint)
        print_set_env_command("OS_AUTH_TYPE", "v3oidcaccesstoken")
        print_set_env_command("OS_IDENTITY_PROVIDER", "egi.eu")
        print_set_env_command("OS_PROTOCOL", protocol)
        print_set_env_command("OS_PROJECT_ID", project_id)
        print_set_env_command("OS_ACCESS_TOKEN", access_token)
    else:
        print(f"VO {vo} not found to have access to site {site}")
    return 1



"""
Implementation of "fedcloud token" commands for interactions with EGI Check-in and
access tokens
"""
import re
import sys
import time
from datetime import datetime

import click
import jwt
import liboidcagent as agent
import requests

from fedcloudclient.decorators import (
    oidc_params,
)

# Minimal lifetime of the access token is 30s and max 24h
_MIN_ACCESS_TOKEN_TIME = 30
_MAX_ACCESS_TOKEN_TIME = 24 * 3600

VO_PATTERN = "urn:mace:egi.eu:group:(.+?):(.+:)*role=member#aai.egi.eu"


def print_error(message, quiet):
    """
    Print error message to stderr if not quiet
    """
    if not quiet:
        print(message, file=sys.stderr)


def decode_token(oidc_access_token):
    """
    Decoding access token to a dict
    :param oidc_access_token:
    :return: dict with token info
    """
    try:
        payload = jwt.decode(oidc_access_token, options={"verify_signature": False})
    except jwt.exceptions.InvalidTokenError:
        print_error("Error: Invalid access token.", False)
        return None
    return payload


def oidc_discover(oidc_url):
    """
    Discover OIDC endpoints

    :param oidc_url: CheckIn URL

    :return: JSON object of OIDC configuration
    """
    request = requests.get(oidc_url + "/.well-known/openid-configuration")
    request.raise_for_status()
    return request.json()


def get_token_from_oidc_agent(oidc_agent_account, quiet=False):
    """
    Get access token from oidc-agent
    :param oidc_agent_account: account name in oidc-agent
    :return: access token, or None on error
    """

    if oidc_agent_account:
        try:
            access_token = agent.get_access_token(
                oidc_agent_account,
                min_valid_period=_MIN_ACCESS_TOKEN_TIME,
                application_hint="fedcloudclient",
            )
            return access_token
        except agent.OidcAgentError as exception:
            print_error(
                "Error getting access token from oidc-agent\n"
                f"Error message: {exception}",
                quiet,
            )
    return None


def get_token_from_mytoken_server(mytoken, mytoken_server, quiet=False):
    """
    Get access token from mytoken server
    :param mytoken:
    :param mytoken_server:
    :return: access token, or None on error
    """

    if mytoken:
        try:
            data = {
                "grant_type": "mytoken",
                "mytoken": mytoken,
            }
            req = requests.post(
                mytoken_server + "/api/v0/token/access",
                json=data,
            )
            req.raise_for_status()
            return req.json().get("access_token")
        except requests.exceptions.HTTPError as exception:
            print_error(
                "Error getting access token from mytoken\n"
                f"Error message: {exception}",
                quiet,
            )
    return None


def check_token(oidc_token, verbose=False):
    """
    Check validity of access token

    :param verbose:
    :param oidc_token: the token to check
    :return: access token, or None on error
    """

    payload = decode_token(oidc_token)
    if payload is None:
        return None

    exp_timestamp = int(payload["exp"])
    current_timestamp = int(time.time())
    exp_time_in_sec = exp_timestamp - current_timestamp

    if exp_time_in_sec < _MIN_ACCESS_TOKEN_TIME:
        print_error("Error: Expired access token.", True)
        return None

    if verbose:
        exp_time_str = datetime.utcfromtimestamp(exp_timestamp).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        print(f"Token is valid until {exp_time_str} UTC")
        if exp_time_in_sec < 24 * 3600:
            print(f"Token expires in {exp_time_in_sec} seconds")
        else:
            exp_time_in_days = exp_time_in_sec // (24 * 3600)
            print(f"Token expires in {exp_time_in_days} days")

    return oidc_token


def get_checkin_id(
    oidc_token,
):
    """
    Get EGI Check-in ID from access token

    :param oidc_token: the token

    :return: Check-in ID
    """
    payload = decode_token(oidc_token)
    if payload is None:
        return None
    return payload["sub"]


def get_access_token(
    oidc_access_token,
    oidc_agent_account,
    mytoken,
    mytoken_server,
):
    """
    Get access token
    Generates new access token from oidc-agent or
    or mytoken

    Check expiration time of access token
    Raise error if no valid token exists

    :param oidc_access_token:
    :param oidc_agent_account:
    :param mytoken:
    :param mytoken_server:
    :return: access token
    """

    access_token = None

    # access token via parameter has the highest priority
    if oidc_access_token:
        access_token = check_token(oidc_access_token)

    # then try to get access token from mytoken server
    if mytoken and access_token is None:
        access_token = get_token_from_mytoken_server(
            mytoken, mytoken_server, quiet=True
        )

    # then, try to get access token from oidc-agent
    if oidc_agent_account and access_token is None:
        access_token = get_token_from_oidc_agent(oidc_agent_account, quiet=True)

    if access_token is None:
        # Nothing available
        raise SystemExit(
            "Error: An access token is needed for the operation. You can specify "
            "access token directly via --oidc-access-token option or use oidc-agent "
            "via --oidc-agent-account or mytoken via --mytoken"
        )

    return access_token


def token_list_vos(oidc_access_token):
    """
    List VO memberships in EGI Check-in

    :param oidc_access_token:

    :return: list of VO names
    """
    oidc_url = decode_token(oidc_access_token)["iss"]
    oidc_ep = oidc_discover(oidc_url)
    request = requests.get(
        oidc_ep["userinfo_endpoint"],
        headers={"Authorization": f"Bearer {oidc_access_token}"},
    )

    request.raise_for_status()
    vos = set()
    pattern = re.compile(VO_PATTERN)
    for claim in request.json().get("eduperson_entitlement", []):
        vo = pattern.match(claim)
        if vo:
            vos.add(vo.groups()[0])
    return sorted(vos)


@click.group()
def token():
    """
    Get details of access token
    """


@token.command()
@oidc_params
def check(access_token):
    """
    Check validity of access token
    """
    check_token(access_token, verbose=True)


@token.command()
@oidc_params
def list_vos(access_token):
    """
    List VO membership(s) of access token
    """
    vos = token_list_vos(access_token)
    print("\n".join(vos))


@token.command()
@oidc_params
def issue(access_token):
    """
    print access token (from mytoken or oidc-agent)
    """
    print(access_token)



"""
Implementation of "fedcloud openstack" or "fedcloud openstack-int" for performing
OpenStack commands on sites
"""

import concurrent.futures
import json
import os
import subprocess  # nosec Subprocess is required for invoking openstack client
import sys
from distutils.spawn import find_executable

import click

from fedcloudclient.decorators import (
    ALL_SITES_KEYWORDS,
    DEFAULT_AUTH_TYPE,
    DEFAULT_IDENTITY_PROVIDER,
    DEFAULT_PROTOCOL,
    all_site_params,
    oidc_params,
    openstack_output_format_params,
    openstack_params,
    site_params,
    vo_params,
)
from fedcloudclient.sites import (
    find_endpoint_and_project_id,
    list_sites,
)

__OPENSTACK_CLIENT = "openstack"
__MAX_WORKER_THREAD = 30
__MISSING_VO_ERROR_CODE = 11

CONFLICTING_ENVS = ["OS_TOKEN", "OS_USER_DOMAIN_NAME"]


def fedcloud_openstack_full(
    oidc_access_token,
    openstack_auth_protocol,
    openstack_auth_type,
    checkin_identity_provider,
    site,
    vo,
    openstack_command,
    json_output=True,
):
    """
    Calling openstack client with full options specified, including support
    for other identity providers and protocols

    :param oidc_access_token: Checkin access token. Passed to openstack client
           as --os-access-token
    :param openstack_auth_protocol: Checkin protocol (openid, oidc). Passed to
           openstack client as --os-protocol
    :param openstack_auth_type: Checkin authentication type (v3oidcaccesstoken).
           Passed to openstack client as --os-auth-type
    :param checkin_identity_provider: Checkin identity provider in mapping (egi.eu).
           Passed to openstack client as --os-identity-provider
    :param site: site ID in GOCDB
    :param vo: VO name
    :param openstack_command: OpenStack command in tuple,
           e.g. ("image", "list", "--long")
    :param json_output: if result is JSON object or string. Default:True

    :return: error code, result or error message
    """

    endpoint, project_id, protocol = find_endpoint_and_project_id(site, vo)
    if endpoint is None:
        return __MISSING_VO_ERROR_CODE, f"VO {vo} not found on site {site}\n"

    if protocol is None:
        protocol = openstack_auth_protocol

    options = (
        "--os-auth-url",
        endpoint,
        "--os-auth-type",
        openstack_auth_type,
        "--os-protocol",
        protocol,
        "--os-identity-provider",
        checkin_identity_provider,
        "--os-access-token",
        oidc_access_token,
    )

    if vo:
        options = options + ("--os-project-id", project_id)

    # Output JSON format is useful for further machine processing
    if json_output:
        options = options + ("--format", "json")

    # Remove conflicting environment
    my_env = os.environ.copy()
    for env in CONFLICTING_ENVS:
        my_env.pop(env, None)

    # Calling openstack client as subprocess, caching stdout/stderr
    # Ignore bandit warning
    completed = subprocess.run(  # nosec
        (__OPENSTACK_CLIENT,) + openstack_command + options,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=my_env,
    )

    error_code = completed.returncode
    error_message = completed.stderr.decode("utf-8")
    result_str = completed.stdout.decode("utf-8")

    if error_code == 0:
        if json_output:
            # Test if openstack command ignore JSON format option
            try:
                json_object = json.loads(result_str)
                return error_code, json_object
            except ValueError:
                return error_code, result_str
        else:
            return error_code, result_str
    else:
        # If error code != 0, return error message instead of result
        # But attach also result string, as applications may print
        # error messages to stdout instead stderr
        return error_code, error_message + result_str


def fedcloud_openstack(
    oidc_access_token, site, vo, openstack_command, json_output=True
):
    """
    Simplified version of fedcloud_openstack_full() using
    default EGI setting for identity provider and protocols
    Calls OpenStack CLI with default options for EGI Check-in

    :param oidc_access_token: Checkin access token.
           Passed to openstack client as --os-access-token
    :param site: site ID in GOCDB
    :param vo: VO name
    :param openstack_command: OpenStack command in tuple,
           e.g. ("image", "list", "--long")
    :param json_output: if result is JSON object or string. Default:True

    :return: error code, result or error message
    """

    return fedcloud_openstack_full(
        oidc_access_token,
        DEFAULT_PROTOCOL,
        DEFAULT_AUTH_TYPE,
        DEFAULT_IDENTITY_PROVIDER,
        site,
        vo,
        openstack_command,
        json_output,
    )


def check_openstack_client_installation():
    """
    Check if openstack command-line client is installed and available via $PATH

    :return: True if available
    """

    return find_executable(__OPENSTACK_CLIENT) is not None


def print_result(
    site,
    vo,
    command,
    exc_msg,
    error_code,
    result,
    json_output,
    ignore_missing_vo,
    first,
):
    """
    Print output from an OpenStack command

    :param site:
    :param vo:
    :param command:
    :param exc_msg:
    :param error_code:
    :param result:
    :param json_output:
    :param ignore_missing_vo:
    :param first:
    :return:
    """

    command = " ".join(command)
    if not json_output:
        if not ignore_missing_vo or (error_code != __MISSING_VO_ERROR_CODE):
            print(f"Site: {site}, VO: {vo}, command: {command}", file=sys.stderr)

        if exc_msg:
            print(f"{site} generated an exception: {exc_msg}")
            print(f"Error message: {result}")

        elif error_code != 0:
            if not ignore_missing_vo or (error_code != __MISSING_VO_ERROR_CODE):
                print("Error code: ", error_code)
                print("Error message: ", result)
        else:
            print(result)
    else:
        site_data = {
            "Site": site,
            "VO": vo,
            "command": command,
            "Exception": exc_msg,
            "Error code": error_code,
            "Result": result,
        }
        separator = "[" if first else ","
        print(separator)
        print(json.dumps(site_data, indent=2))


@click.command(context_settings={"ignore_unknown_options": True})
@all_site_params
@vo_params
@oidc_params
@openstack_output_format_params
@openstack_params
@click.argument("openstack_command", required=True, nargs=-1)
def openstack(
    access_token,
    site,
    all_sites,
    vo,
    openstack_command,
    openstack_auth_protocol,
    openstack_auth_type,
    openstack_auth_provider,
    ignore_missing_vo,
    json_output,
):
    """
    Execute OpenStack commands on site and VO
    """

    if not check_openstack_client_installation():
        raise SystemExit('Error: OpenStack command-line client "openstack" not found')

    if site in ALL_SITES_KEYWORDS or all_sites:
        sites = list_sites()
    else:
        sites = [site]

    # Multi-thread execution of OpenStack commands
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=__MAX_WORKER_THREAD
    ) as executor:
        # Start OpenStack operation with each site
        results = {
            executor.submit(
                fedcloud_openstack_full,
                access_token,
                openstack_auth_protocol,
                openstack_auth_type,
                openstack_auth_provider,
                site,
                vo,
                openstack_command,
                json_output,
            ): site
            for site in sites
        }

        # Get results and print them
        first = True

        # Get the result, first come first serve
        for future in concurrent.futures.as_completed(results):
            site = results[future]
            exc_msg = None
            try:
                error_code, result = future.result()
            except Exception as exc:
                exc_msg = exc

            # Print result
            print_result(
                site,
                vo,
                openstack_command,
                exc_msg,
                error_code,
                result,
                json_output,
                ignore_missing_vo,
                first,
            )
            first = False

        # Print list enclosing ']' for JSON
        if json_output:
            print("]")


@click.command()
@site_params
@vo_params
@oidc_params
@openstack_params
def openstack_int(
    access_token,
    site,
    vo,
    openstack_auth_protocol,
    openstack_auth_type,
    openstack_auth_provider,
):
    """
    Interactive OpenStack client on site and VO
    """

    if not check_openstack_client_installation():
        raise SystemExit('Error: OpenStack command-line client "openstack" not found')

    endpoint, project_id, protocol = find_endpoint_and_project_id(site, vo)
    if endpoint is None:
        raise SystemExit(f"Error: VO {vo} not found on site {site}")

    if protocol is None:
        protocol = openstack_auth_protocol
    my_env = os.environ.copy()
    my_env["OS_AUTH_URL"] = endpoint
    my_env["OS_AUTH_TYPE"] = openstack_auth_type
    my_env["OS_PROTOCOL"] = protocol
    my_env["OS_IDENTITY_PROVIDER"] = openstack_auth_provider
    my_env["OS_ACCESS_TOKEN"] = access_token
    my_env["OS_PROJECT_ID"] = project_id

    # Calling OpenStack client as subprocess
    # Ignore bandit warning
    subprocess.run(__OPENSTACK_CLIENT, env=my_env)  # nosec



"""
"fedcloud endpoint" commands are complementary part of the "fedcloud site" commands.

Instead of using site configurations defined in files saved in GitHub repository or
local disk, the commands try to get site information directly from GOCDB
(Grid Operations Configuration Management Database) https://goc.egi.eu/ or make probe
test on sites
"""

import os
import sys
from urllib import parse

import click
import requests
from defusedxml import ElementTree
from tabulate import tabulate

from fedcloudclient.decorators import (
    ALL_SITES_KEYWORDS,
    all_site_params,
    oidc_params,
    project_id_params,
    site_params,
)
from fedcloudclient.shell import print_set_env_command
from fedcloudclient.sites import find_vo_from_project_id

OPENSTACK_NOVA = "org.openstack.nova"
GOCDB_PUBLICURL = "https://goc.egi.eu/gocdbpi/public/"
TIMEOUT = 10
TLS_ERROR_MSG = "TLS error. IGTF certificates may not have been installed"
CONNECTION_ERROR_MSG = "Connection error"
SCOPED_TOKEN_ERROR_MSG = "Unable to get a scoped token"
UNSCOPED_TOKEN_ERROR_MSG = "Unable to get an unscoped token"


class TokenException(Exception):
    """Exception for Token related errors"""

    # Basic exception. This should be extended


def get_sites():
    """
    Get list of sites (using GOCDB instead of site configuration)

    :return: list of site IDs
    """
    query = {"method": "get_site_list", "certification_status": "Certified"}
    url = "?".join([GOCDB_PUBLICURL, parse.urlencode(query)])
    request = requests.get(url)
    sites = []
    if request.status_code == 200:
        root = ElementTree.fromstring(request.text)
        for site_object in root:
            sites.append(site_object.attrib.get("NAME"))
    else:
        print("Something went wrong...")
        print(request.status_code)
        print(request.text)
    return sites


def find_endpoint(service_type, production=True, monitored=True, site=None):
    """
    Searching GOCDB for endpoints according to service types and status

    :param service_type:
    :param production:
    :param monitored:
    :param site: list of sites, None for searching all sites

    :return: list of endpoints
    """
    query = {"method": "get_service_endpoint", "service_type": service_type}
    if monitored:
        query["monitored"] = "Y"
    if site:
        query["sitename"] = site
        sites = [site]
    else:
        sites = get_sites()
    url = "?".join([GOCDB_PUBLICURL, parse.urlencode(query)])
    request = requests.get(url)
    endpoints = []
    if request.status_code == 200:
        root = ElementTree.fromstring(request.text)
        for site_object in root:
            if production:
                prod = site_object.find("IN_PRODUCTION").text.upper()
                if prod != "Y":
                    continue
            os_url = site_object.find("URL").text
            ep_site = site_object.find("SITENAME").text
            if ep_site not in sites:
                continue
            # os_url = urlparse.urlparse(sp.find('URL').text)
            # sites[sp.find('SITENAME').text] = urlparse.urlunparse(
            #    (os_url[0], os_url[1], os_url[2], '', '', ''))
            endpoints.append([site_object.find("SITENAME").text, service_type, os_url])
    else:
        print("Something went wrong...")
        print(request.status_code)
        print(request.text)
    return endpoints


def get_keystone_url(os_auth_url, path):
    """
    Helper function for fixing Keystone URL
    """
    url = parse.urlparse(os_auth_url)
    prefix = url.path.rstrip("/")
    if prefix.endswith("v2.0") or prefix.endswith("v3"):
        prefix = os.path.dirname(prefix)
    path = os.path.join(prefix, path)
    return parse.urlunparse((url[0], url[1], path, url[3], url[4], url[5]))


def get_unscoped_token(os_auth_url, access_token):
    """
    Get an unscoped token, will try all protocols if needed
    """
    protocols = ["openid", "oidc"]
    for protocol in protocols:
        try:
            unscoped_token = retrieve_unscoped_token(
                os_auth_url, access_token, protocol
            )
            return unscoped_token, protocol
        except RuntimeError:
            pass
    raise TokenException(SCOPED_TOKEN_ERROR_MSG)


def get_scoped_token(os_auth_url, access_token, project_id):
    """
    Get a scoped token, will try all protocols if needed
    """
    unscoped_token, protocol = get_unscoped_token(os_auth_url, access_token)
    url = get_keystone_url(os_auth_url, "/v3/auth/tokens")
    body = {
        "auth": {
            "identity": {"methods": ["token"], "token": {"id": unscoped_token}},
            "scope": {"project": {"id": project_id}},
        }
    }
    request = requests.post(url, json=body)
    # pylint: disable=no-member
    if request.status_code != requests.codes.created:
        raise TokenException(SCOPED_TOKEN_ERROR_MSG)

    return request.headers["X-Subject-Token"], protocol


def retrieve_unscoped_token(os_auth_url, access_token, protocol="openid"):
    """
    Request an unscoped token
    """
    url = get_keystone_url(
        os_auth_url,
        f"/v3/OS-FEDERATION/identity_providers/egi.eu/protocols/{protocol}/auth",
    )
    request = requests.post(
        url,
        headers={"Authorization": f"Bearer {access_token}"},
        timeout=TIMEOUT,
    )
    # pylint: disable=no-member
    if request.status_code != requests.codes.created:
        raise TokenException(UNSCOPED_TOKEN_ERROR_MSG)

    return request.headers["X-Subject-Token"]


def get_projects_from_single_site(os_auth_url, unscoped_token):
    """
    Get list of projects from unscoped token
    """
    url = get_keystone_url(os_auth_url, "/v3/auth/projects")
    request = requests.get(url, headers={"X-Auth-Token": unscoped_token})
    request.raise_for_status()
    return request.json()["projects"]


def format_project_as_list(site_name, project):
    """
    Format project data as a list
    """
    return [[project["id"], project["name"], project["enabled"], site_name]]


def format_project_as_dict(site_name, project):
    """
    Format project data as a dictionary
    """
    return [
        {
            "project_id": project["id"],
            "name": project["name"],
            "enabled": project["enabled"],
            "site": site_name,
        }
    ]


def get_projects_from_sites(access_token, site):
    """
    Get all projects from site(s) using access token, in the default output format (list)
    """
    return get_projects_from_sites_as_list(access_token, site)


def get_projects_from_sites_as_list(access_token, site):
    """
    Get all projects from site(s) using access token, as a list
    """
    return get_projects_from_sites_with_format(
        access_token, site, format_project_as_list
    )


def get_projects_from_sites_as_dict(access_token, site):
    """
    Get all projects from site(s) using access token, as a dictionary
    """
    return get_projects_from_sites_with_format(
        access_token, site, format_project_as_dict
    )


def get_projects_from_sites_with_format(access_token, site, output_format_function):
    """
    Get all projects from site(s) using access token
    """
    project_list = []
    project_error_list = []
    for site_ep in find_endpoint(OPENSTACK_NOVA, site=site):
        site_name = site_ep[0]
        site_os_auth_url = site_ep[2]

        try:
            unscoped_token, _ = get_unscoped_token(site_os_auth_url, access_token)
            for project in get_projects_from_single_site(
                site_os_auth_url, unscoped_token
            ):
                project_list.extend(output_format_function(site_name, project))
        except TokenException:
            # e.g. The user may have no permissions
            pass
        except requests.exceptions.SSLError:
            project_error_list.append([site_name, TLS_ERROR_MSG])
        except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
            # The site is unavailable e.g. HTTP 500
            project_error_list.append([site_name, CONNECTION_ERROR_MSG])
        except RuntimeError:
            # Logging needed
            pass
    return project_list, project_error_list


@click.group()
def endpoint():
    """
    Obtain endpoint details and scoped tokens
    """


@endpoint.command()
@all_site_params
@oidc_params
def vos(
    access_token,
    site,
    all_sites,
):
    """
    List projects from site(s)
    """
    if site in ALL_SITES_KEYWORDS or all_sites:
        site = None

    project_list, project_error_list = get_projects_from_sites(access_token, site)
    if len(project_list) > 0:
        for p in project_list:
            vo = find_vo_from_project_id(p[3], p[0])
            p.insert(0, vo)
        print(
            tabulate(
                project_list, headers=["VO", "id", "Project name", "enabled", "site"]
            )
        )
    else:
        print(f"Error: You probably do not have access to any project at site {site}")
        print(
            'Check your access token and VO memberships using "fedcloud token list-vos"'
        )
    if len(project_error_list) > 0:
        print("[WARN] Sites not available: ", project_error_list, file=sys.stderr)


@endpoint.command()
@all_site_params
@oidc_params
def projects(
    access_token,
    site,
    all_sites,
):
    """
    List projects from site(s)
    """
    if site in ALL_SITES_KEYWORDS or all_sites:
        site = None

    project_list, project_error_list = get_projects_from_sites(access_token, site)
    if len(project_list) > 0:
        print(tabulate(project_list, headers=["id", "Name", "enabled", "site"]))
    else:
        print(f"Error: You probably do not have access to any project at site {site}")
        print(
            'Check your access token and VO memberships using "fedcloud token list-vos"'
        )
    if len(project_error_list) > 0:
        print("[WARN] Sites not available: ", project_error_list, file=sys.stderr)


@endpoint.command()
@site_params
@project_id_params
@oidc_params
def token(
    access_token,
    site,
    project_id,
):
    """
    Get scoped Keystone token for site and project
    """
    if site in ALL_SITES_KEYWORDS:
        print("Cannot get tokens for ALL_SITES")
        raise click.Abort()

    # Getting sites from GOCDB
    # assume first one is ok
    site_ep = find_endpoint(OPENSTACK_NOVA, site=site).pop()
    os_auth_url = site_ep[2]
    try:
        scoped_token, _ = get_scoped_token(os_auth_url, access_token, project_id)
        print_set_env_command("OS_TOKEN", scoped_token)
    except RuntimeError:
        print(f"Error: Unable to get Keystone token from site {site}")


@endpoint.command("list")
@all_site_params
@click.option(
    "--service-type",
    default=OPENSTACK_NOVA,
    help="Service type in GOCDB",
    show_default=True,
)
@click.option(
    "--production/--not-production",
    default=True,
    help="Production status",
    show_default=True,
)
@click.option(
    "--monitored/--not-monitored",
    default=True,
    help="Monitoring status",
    show_default=True,
)
def list_(service_type, production, monitored, site, all_sites):
    """
    List endpoints in site(s), will query GOCDB
    """
    if site in ALL_SITES_KEYWORDS or all_sites:
        site = None

    endpoints = find_endpoint(service_type, production, monitored, site)
    print(tabulate(endpoints, headers=["Site", "type", "URL"]))


@endpoint.command()
@site_params
@project_id_params
@oidc_params
def env(
    access_token,
    site,
    project_id,
):
    """
    Generate OS environment variables for site and project
    """
    if site in ALL_SITES_KEYWORDS:
        print("Cannot generate environment variables for ALL_SITES")
        raise click.Abort()

    # Get the right endpoint from GOCDB
    # assume first one is ok
    site_ep = find_endpoint(OPENSTACK_NOVA, site=site).pop()
    os_auth_url = site_ep[2]

    try:
        _, protocol = get_scoped_token(os_auth_url, access_token, project_id)
        print(f"# environment for {site}")
        print_set_env_command("OS_PROJECT_ID", project_id)
        print_set_env_command("OS_AUTH_URL", os_auth_url)
        print_set_env_command("OS_AUTH_TYPE", "v3oidcaccesstoken")
        print_set_env_command("OS_IDENTITY_PROVIDER", "egi.eu")
        print_set_env_command("OS_PROTOCOL", protocol)
        print_set_env_command("OS_ACCESS_TOKEN", access_token)
    except RuntimeError as runtime_error:
        raise SystemExit(
            f"Error: Unable to get Keystone token from site {site}"
        ) from runtime_error



"""
Support shell functions
"""

import os
import re
import sys
from enum import Enum

from psutil import Process


class Shell(Enum):
    """
    Shell types on different OSes
    """

    LINUX = 1
    WINDOWS_COMMAND_PROMPT = 2
    POWER_SHELL = 3


def get_shell_type():
    """
    Check if running on Windows and what shell type were we launched from
    """
    if sys.platform.startswith("win"):
        parent_proc = os.getppid()
        parent_name = Process(parent_proc).name()

        if bool(re.match("pwsh*|pwsh.exe|powershell.exe", parent_name)):
            return Shell.POWER_SHELL

        return Shell.WINDOWS_COMMAND_PROMPT

    return Shell.LINUX


def print_set_env_command(name, value):
    """
    Print command to set environment variable,
    format it correctly for the current platform
    """
    shell_type = get_shell_type()
    if shell_type == Shell.LINUX:
        print(f'export {name!s}="{value!s}";')
    elif shell_type == Shell.POWER_SHELL:
        print(f'$Env:{name!s}="{value!s}";')
    else:
        print(f"set {name!s}={value!s}")


def print_comment(comment):
    """
    Print comment command,
    format it correctly for the current platform
    """
    shell_type = get_shell_type()
    if shell_type == Shell.LINUX:
        print(f"# {comment!s}")
    elif shell_type == Shell.POWER_SHELL:
        print(f"# {comment!s}")
    else:
        print(f"rem {comment!s}")



"""
Decorators for command-line parameters
"""
from functools import wraps

import click
from click_option_group import (
    RequiredAnyOptionGroup,
    RequiredMutuallyExclusiveOptionGroup,
    optgroup,
)

DEFAULT_MYTOKEN_SERVER = "https://mytoken.data.kit.edu"
DEFAULT_PROTOCOL = "openid"
DEFAULT_AUTH_TYPE = "v3oidcaccesstoken"
DEFAULT_IDENTITY_PROVIDER = "egi.eu"

ALL_SITES_KEYWORDS = {"ALL_SITES", "ALL-SITES"}

# Decorator for --oidc-access-token
oidc_access_token_params = click.option(
    "--oidc-access-token",
    help="OIDC access token",
    envvar="OIDC_ACCESS_TOKEN",
    metavar="token",
)

# Decorator for --site
site_params = click.option(
    "--site",
    help="Name of the site",
    required=True,
    envvar="EGI_SITE",
    metavar="site-name",
)


def all_site_params(func):
    """
    Decorator for all-sites options
    """

    @optgroup.group(
        "Site",
        cls=RequiredMutuallyExclusiveOptionGroup,
        help="Single Openstack site or all sites",
    )
    @optgroup.option(
        "--site",
        help="Name of the site or ALL_SITES",
        envvar="EGI_SITE",
        metavar="site-name",
    )
    @optgroup.option(
        "--all-sites",
        "-a",
        help="Short option for all sites (equivalent --site ALL_SITES)",
        is_flag=True,
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


# Decorator for --project-id
project_id_params = click.option(
    "--project-id",
    help="Project ID",
    required=True,
    envvar="OS_PROJECT_ID",
    metavar="project-id",
)

# Decorator for --auth-file (used in ec3 module)
auth_file_params = click.option(
    "--auth-file",
    help="Authorization file",
    default="auth.dat",
    show_default=True,
    metavar="auth-file",
)

# Decorator for --vo
vo_params = click.option(
    "--vo",
    help="Name of the VO",
    required=True,
    envvar="EGI_VO",
    metavar="vo-name",
)


def site_vo_params(func):
    """
    Decorator for site and VO parameters
    """

    @site_params
    @vo_params
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def oidc_params(func):
    """
    Decorator for OIDC parameters.
    Get access token from oidc-* parameters and replace them in the wrapper function
    """

    @optgroup.group("OIDC token", help="Choose one of options for providing token")
    @optgroup.option(
        "--oidc-agent-account",
        help="Account name in oidc-agent",
        envvar="OIDC_AGENT_ACCOUNT",
        metavar="account",
    )
    @optgroup.option(
        "--oidc-access-token",
        help="OIDC access token",
        envvar="OIDC_ACCESS_TOKEN",
        metavar="token",
    )
    @optgroup.option(
        "--mytoken",
        help="Mytoken string",
        envvar="FEDCLOUD_MYTOKEN",
        metavar="mytoken",
    )
    @optgroup.option(
        "--mytoken-server",
        help="Mytoken sever",
        envvar="FEDCLOUD_MYTOKEN_SERVER",
        default=DEFAULT_MYTOKEN_SERVER,
        show_default=True,
        metavar="mytoken-server",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        from fedcloudclient.checkin import get_access_token

        access_token = get_access_token(
            kwargs.pop("oidc_access_token"),
            kwargs.pop("oidc_agent_account"),
            kwargs.pop("mytoken"),
            kwargs.pop("mytoken_server"),
        )
        kwargs["access_token"] = access_token
        return func(*args, **kwargs)

    return wrapper


def openstack_params(func):
    """
    Decorator for OpenStack authentication parameters
    """

    @optgroup.group(
        "OpenStack authentication",
        help="Only change default values if necessary",
    )
    @optgroup.option(
        "--openstack-auth-protocol",
        help="Authentication protocol",
        envvar="OPENSTACK_AUTH_PROTOCOL",
        default=DEFAULT_PROTOCOL,
        show_default=True,
        metavar="",
    )
    @optgroup.option(
        "--openstack-auth-provider",
        help="Identity provider",
        envvar="OPENSTACK_AUTH_PROVIDER",
        default=DEFAULT_IDENTITY_PROVIDER,
        show_default=True,
        metavar="",
    )
    @optgroup.option(
        "--openstack-auth-type",
        help="Authentication type",
        envvar="OPENSTACK_AUTH_TYPE",
        default=DEFAULT_AUTH_TYPE,
        show_default=True,
        metavar="",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def openstack_output_format_params(func):
    """
    Decorator for output format options for OpenStack
    """

    @optgroup.group(
        "Output format",
        help="Parameters for formatting outputs",
    )
    @optgroup.option(
        "--ignore-missing-vo",
        "-i",
        help="Ignore sites that do not support the VO",
        is_flag=True,
    )
    @optgroup.option(
        "--json-output",
        "-j",
        help="Print output as JSON object",
        is_flag=True,
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def flavor_specs_params(func):
    """
    Decorator for flavor specs
    """

    @optgroup.group(
        "Flavor specs options",
        cls=RequiredAnyOptionGroup,
        help="Parameters for flavor specification",
    )
    @optgroup.option(
        "--flavor-specs",
        help=(
            "Flavor specifications, e.g. 'VCPUs==2' or 'Disk>100'."
            " May be specified more times, or joined, e.g. 'VCPUs==2 & RAM>2048'"
        ),
        multiple=True,
        metavar="flavor-specs",
    )
    @optgroup.option(
        "--vcpus",
        help="Number of VCPUs (equivalent --flavor-specs VCPUs==vcpus)",
        metavar="vcpus",
    )
    @optgroup.option(
        "--RAM",
        help="Amount of RAM in MB (equivalent --flavor-specs RAM==memory)",
        metavar="memory",
    )
    @optgroup.option(
        "--gpus",
        help="Number of GPUs (equivalent --flavor-specs Properties.Accelerator:Number==gpus)",
        metavar="gpus",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def image_specs_params(func):
    """
    Decorator for flavor specs
    """

    @optgroup.group(
        "Image specs options",
        help="Parameters for image specification",
    )
    @optgroup.option(
        "--image-specs",
        required=True,
        help="Image specifications, e.g. 'Name=~\"20.04\"' or 'Name=~\"CentOS 8\"'",
        multiple=True,
        metavar="image-specs",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def network_specs_params(func):
    """
    Decorator for network specs
    """

    @optgroup.group(
        "Network specs options",
        help="Parameters for network specification",
    )
    @optgroup.option(
        "--network-specs",
        help="Network specifications: 'default', 'public', 'private'",
        type=click.Choice(["default", "public", "private"], case_sensitive=False),
        default="default",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def select_output_format_params(func):
    """
    Decorator for option for printing outputs from searching
    """

    @optgroup.group(
        "Output format",
        help="Parameters for formatting outputs",
    )
    @optgroup.option(
        "--output-format",
        help=(
            "Option for printing matched results, 'first' for printing only best matched item,"
            " 'list' for printing all matched resources, and 'YAML' or 'JSON' for full output"
        ),
        type=click.Choice(["first", "list", "YAML", "JSON"], case_sensitive=False),
        default="JSON",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper



"""
Main CLI module
"""

import click

from fedcloudclient.checkin import token
from fedcloudclient.ec3 import ec3
from fedcloudclient.endpoint import endpoint
from fedcloudclient.openstack import openstack, openstack_int
from fedcloudclient.secret import secret
from fedcloudclient.select import select
from fedcloudclient.sites import site


@click.group()
@click.version_option()
def cli():
    """
    CLI main function. Intentionally empty
    """


cli.add_command(token)
cli.add_command(endpoint)
cli.add_command(ec3)
cli.add_command(site)
cli.add_command(secret)
cli.add_command(select)
cli.add_command(openstack)
cli.add_command(openstack_int)


if __name__ == "__main__":
    cli()



"""
Implementation of "fedcloud select" commands for selecting suitable flavors
or images on different OpenStack sites
"""
import json

import click
import yaml
from jsonpath_ng.exceptions import JSONPathError
from jsonpath_ng.ext import parse

from fedcloudclient.decorators import (
    flavor_specs_params,
    image_specs_params,
    network_specs_params,
    oidc_params,
    select_output_format_params,
    site_params,
    vo_params,
)
from fedcloudclient.openstack import fedcloud_openstack
from fedcloudclient.sites import find_endpoint_and_project_id

FILTER_TEMPLATE = "$[?( {specs} )]"
GET_FLAVOR_COMMAND = ("flavor", "list", "--long")
GET_NETWORK_COMMAND = ("network", "list", "--long")
GET_IMAGE_COMMAND = ("image", "list", "--long", "--sort", "created_at:desc")


def compare_flavors(flavor_item):
    """
    Helper function for sorting flavors.
    Sorting order: Flavors with lower resources first.
    Resource importance order: GPUs, CPUs, RAM, Disk, Ephemeral.

    :param flavor_item:

    :return:
    """
    return (
        flavor_item.get("Properties", {}).get("Accelerator:Number", "0"),
        flavor_item.get("VCPUs"),
        flavor_item.get("RAM"),
        flavor_item.get("Disk"),
        flavor_item.get("Ephemeral"),
    )


def sort_flavors(flavors):
    """
    Sorting flavors, flavors with low resources first

    :param flavors: flavor list in JSON

    :return: sorted flavor list in JSON
    """
    return sorted(flavors, key=compare_flavors)


def get_resource(oidc_access_token, site, vo, command):
    """
    Getting list of flavors for the given site and VO, using
    `openstack flavor list --long` command

    :param command:
    :param site:
    :param oidc_access_token:
    :param vo: self descriptive

    :return: list of flavors in JSON formats
    """

    error_code, resource = fedcloud_openstack(oidc_access_token, site, vo, command)
    if error_code:
        command_string = " ".join(command)
        raise SystemExit(
            "Error during executing command: "
            f"fedcloud openstack {command_string} --site {site} --vo {vo}\n"
            f"Error code: {error_code}\n"
            f"Error message: {resource}\n"
        )
    return resource


def construct_filter(specs, match_template):
    """
    Build parser according to specification

    :param match_template:
    :param specs: list of strings of specifications

    :return: filter string
    """

    spec_string = " & ".join(specs)
    return match_template.format(specs=spec_string)


def get_parser(filter_string):
    """
    Build parser according to specification.
    Raise SystemExit with error message if invalid filter string

    :param filter_string:

    :return: jsonpath parser
    """

    try:
        parser = parse(filter_string)
    except JSONPathError as exception:
        raise SystemExit(
            "Error during constructing filter\n"
            f"Filter string: {filter_string}\n"
            f"{exception}"
        )
    return parser


def do_filter(parser, input_list):
    """
    Filter flavor by applying  parser on the flavor list

    :param input_list: list of input data in JSON format
    :param parser: jsonpath parser

    :return: list of matched flavors in JSON
    """
    try:
        matched = [match.value for match in parser.find(input_list)]
    except TypeError as exception:
        raise SystemExit(
            "TypeError during filtering result\n"
            "Probably string value for numeric property in filter\n"
            f"{exception}"
        )
    return matched


def compare_network(network_item):
    """
    Helper function for sorting network
    Sorting order: public, private shared, private

    :param network_item:

    :return:
    """
    return (
        1 if network_item.get("Router Type") else 0,
        1 if network_item.get("Shared") else 0,
    )


def filter_network(networks, network_specs, project_id):
    """
    Filter network according to network specifications

    :param networks: List of all networks
    :param network_specs: Network specifications
    :param project_id: Project ID for assessment of access

    :return: List of matched network
    """
    match_networks = []

    for network_item in networks:
        if network_item.get("Shared") or network_item.get("Project") == project_id:
            if network_specs == "private":
                if not network_item.get("Router Type"):
                    match_networks.append(network_item)
            elif network_specs == "public":
                if network_item.get("Router Type"):
                    match_networks.append(network_item)
            else:
                match_networks.append(network_item)

    return sorted(match_networks, key=compare_network, reverse=True)


def print_output(resource_list, output_format):
    """
    Print resource list according to output format

    :param resource_list: resource list
    :param output_format: output format

    :return: None
    """

    if len(resource_list) == 0:
        raise SystemExit("Error: No resource matched specification")

    if output_format == "first":
        print(resource_list[0].get("Name"))
    elif output_format == "list":
        for resource_item in resource_list:
            print(resource_item.get("Name"))
    elif output_format == "YAML":
        print(yaml.dump(resource_list, sort_keys=False))
    else:
        print(json.dumps(resource_list, indent=4))


@click.group()
def select():
    """
    Select resources according to specification
    """


@select.command()
@site_params
@vo_params
@flavor_specs_params
@select_output_format_params
@oidc_params
def flavor(
    site,
    vo,
    access_token,
    flavor_specs,
    vcpus,
    ram,
    gpus,
    output_format,
):
    """
    Select suitable flavors according to flavor specifications on the given site/VO
    """

    flavor_specs = list(flavor_specs)
    if vcpus:
        flavor_specs.append(f"VCPUs=={vcpus}")
    if ram:
        flavor_specs.append(f"=={ram}")
    if gpus:
        flavor_specs.append(f"Properties.'Accelerator:Number'=~'{gpus}'")

    filter_string = construct_filter(flavor_specs, FILTER_TEMPLATE)
    parser = get_parser(filter_string)
    flavors = get_resource(access_token, site, vo, GET_FLAVOR_COMMAND)

    match_flavors = do_filter(parser, flavors)

    sorted_flavors = sort_flavors(match_flavors)
    print_output(sorted_flavors, output_format)


@select.command()
@site_params
@vo_params
@image_specs_params
@select_output_format_params
@oidc_params
def image(
    site,
    vo,
    access_token,
    image_specs,
    output_format,
):
    """
    Select suitable images according to image specifications on the given site/VO
    """

    image_specs = list(image_specs)
    filter_string = construct_filter(image_specs, FILTER_TEMPLATE)
    parser = get_parser(filter_string)
    images = get_resource(access_token, site, vo, GET_IMAGE_COMMAND)

    match_images = do_filter(parser, images)

    print_output(match_images, output_format)


@select.command()
@site_params
@vo_params
@network_specs_params
@select_output_format_params
@oidc_params
def network(
    site,
    vo,
    access_token,
    network_specs,
    output_format,
):
    """
    Select suitable network according to network specifications on the given site/VO
    """

    networks = get_resource(access_token, site, vo, GET_NETWORK_COMMAND)

    _, project_id, _ = find_endpoint_and_project_id(site, vo)
    match_network = filter_network(networks, network_specs, project_id)

    print_output(match_network, output_format)



"""
Implementation of ec3 commands for deploying EC3 (Elastic Cloud Computing
Cluster) in Cloud via Infrastructure Manager
"""

import os
import time

import click
import jwt

from fedcloudclient.decorators import (
    ALL_SITES_KEYWORDS,
    auth_file_params,
    oidc_params,
)
from fedcloudclient.sites import find_endpoint_and_project_id, site_vo_params

__MIN_EXPIRATION_TIME = 300

EC3_REFRESHTOKEN_TEMPLATE = """
description refreshtoken (
    kind = 'component' and
    short = 'Tool to refresh LToS access token.' and
    content = 'Tool to refresh LToS access token.'
)
configure front (
@begin
  - vars:
      CLIENT_ID: %(client_id)s
      CLIENT_SECRET: %(client_secret)s
      REFRESH_TOKEN: %(refresh_token)s
    tasks:
    - name: Check if docker is available
      command: which docker
      changed_when: false
      failed_when: docker_installed.rc not in [0,1]
      register: docker_installed
    - name: local install of fedcloudclient
      block:
      - name: Create dir /usr/local/ec3/
        file: path=/usr/local/ec3/ state=directory
      - name: install git
        package:
          name: git
          state: present
      - name: install fedcloudclient
        pip:
          name:
          - fedcloudclient
      - cron:
          name: "refresh token"
          minute: "*/5"
          job: "[ -f /usr/local/ec3/auth.dat ]
            && /usr/local/bin/fedcloudclient endpoint ec3-refresh
               --oidc-client-id {{ CLIENT_ID }}
               --oidc-client-secret {{ CLIENT_SECRET }}
               --oidc-refresh-token {{ REFRESH_TOKEN }}
               --auth-file /usr/local/ec3/auth.dat &> /var/log/refresh.log"
          user: root
          cron_file: refresh_token
          state: present
      when: docker_installed.rc not in [ 0 ]
    - name: local install of fedcloudclient
      block:
      - cron:
          name: "refresh token"
          minute: "*/5"
          job: "[ -f /usr/local/ec3/auth.dat ]
            && docker run -v /usr/local/ec3/auth.dat:/usr/local/ec3/auth.dat
               tdviet/fedcloudclient fedcloudclient endpoint ec3-refresh
               --oidc-client-id {{ CLIENT_ID }}
               --oidc-client-secret {{ CLIENT_SECRET }}
               --oidc-refresh-token {{ REFRESH_TOKEN }}
               --auth-file /usr/local/ec3/auth.dat &> /var/log/refresh.log"
          user: root
          cron_file: refresh_token
          state: present
      when: docker_installed.rc not in [ 1 ]
@end
)
"""


@click.group()
def ec3():
    """
    EC3 cluster provisioning
    """


@ec3.command()
@oidc_params
@auth_file_params
def refresh(
    access_token,
    auth_file,
):
    """
    Refresh token in EC3 authorization file
    """
    # Get the right endpoint from GOCDB
    auth_file_contents = []
    with open(auth_file, "r") as file:
        for raw_line in file.readlines():
            line = raw_line.strip()
            if "OpenStack" in line:
                auth_tokens = []
                for token in line.split(";"):
                    if token.strip().startswith("password"):
                        current_access_token = token.split("=")[1].strip()
                        if current_access_token[0] in ["'", '"']:
                            current_access_token = current_access_token[1:-1]
                        # FIXME(enolfc): add verification
                        payload = jwt.decode(
                            current_access_token, options={"verify_signature": False}
                        )
                        now = int(time.time())
                        expires = int(payload["exp"])
                        if expires - now < __MIN_EXPIRATION_TIME:
                            current_access_token = access_token
                        auth_tokens.append(f"password = {current_access_token}")
                    else:
                        auth_tokens.append(token.strip())
                auth_file_contents.append("; ".join(auth_tokens))
            elif line:
                auth_file_contents.append(line)

    with open(auth_file, "w+") as file:
        file.write("\n".join(auth_file_contents))


@ec3.command()
@site_vo_params
@oidc_params
@auth_file_params
@click.option(
    "--template-dir",
    help="EC3 templates dir",
    default="./templates",
    show_default=True,
)
@click.option("--force", is_flag=True, help="Force rewrite of files")
def init(
    access_token,
    site,
    vo,
    auth_file,
    template_dir,
    force,
):
    """
    Create EC3 authorization file and template
    """
    if os.path.exists(auth_file) and not force:
        print(
            "Auth file already exists, not replacing unless --force option is included"
        )
        raise click.Abort()

    if site in ALL_SITES_KEYWORDS:
        print("EC3 commands cannot be used with ALL_SITES")
        raise click.Abort()

    endpoint, project_id, protocol = find_endpoint_and_project_id(site, vo)
    site_auth = [
        f"id = {site}",
        "type = OpenStack",
        "username = egi.eu",
        f"tenant = {protocol}",
        "auth_version = 3.x_oidc_access_token",
        f"host = {endpoint}",
        f"domain = {project_id}",
        f"password = {access_token}",
    ]
    auth_file_contents = [";".join(site_auth)]

    if os.path.exists(auth_file):
        with open(auth_file, "r") as file:
            for line in file.readlines():
                if "OpenStack" in line:
                    continue
                auth_file_contents.append(line)

    with open(auth_file, "w+") as file:
        file.write("\n".join(auth_file_contents))

    if not os.path.exists(template_dir):
        os.mkdir(template_dir)

    # FIXME: this should not be used at all!
    with open(os.path.join(template_dir, "refresh.radl"), "w+") as file:
        token = dict(  # nosec
            client_id="ADD_CLIENT_ID_HERE",
            client_secret="ADD_CLIENT_SECRET_HERE",
            refresh_token="ADD_REFRESH_TOKEN_HERE",
        )
        file.write(EC3_REFRESHTOKEN_TEMPLATE % token)
